"""add circuit.runtime_variant column

Revision ID: 825d283e0743
Revises: 0dd436073e84
Create Date: 2024-11-11 17:26:43.981090

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "825d283e0743"
down_revision = "0dd436073e84"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("circuits", sa.Column("runtime_variant", sa.VARCHAR(64), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("circuits", "runtime_variant")
    # ### end Alembic commands ###
