"""rename misrepresentative column names

Revision ID: 6e38326a056a
Revises: e81f833e28fa
Create Date: 2024-03-09 13:26:31.168685

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "6e38326a056a"
down_revision = "e81f833e28fa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("workers", "available_circuits", new_column_name="available_slots")
    op.alter_column("workers", "occupied_circuits", new_column_name="occupied_slots")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("workers", "available_slots", new_column_name="available_circuits")
    op.alter_column("workers", "occupied_slots", new_column_name="occupied_circuits")
    # ### end Alembic commands ###
