"""add worker.wildcard_traffic_port column

Revision ID: a06b7ba8f4bd
Revises: 6e38326a056a
Create Date: 2024-10-22 10:25:40.370766

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a06b7ba8f4bd"
down_revision = "6e38326a056a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workers", sa.Column("wildcard_traffic_port", sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workers", "wildcard_traffic_port")
    # ### end Alembic commands ###
