"""add worker.traefik_last_used_marker_path

Revision ID: 66f87e010f90
Revises: 825d283e0743
Create Date: 2025-02-25 12:31:36.445094

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "66f87e010f90"
down_revision = "825d283e0743"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "workers", sa.Column("traefik_last_used_marker_path", sa.String(length=1024), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workers", "traefik_last_used_marker_path")
    # ### end Alembic commands ###
